"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from .utils import (
    Weekly as Weekly,
    Monthly as Monthly,
    Yearly as Yearly,
    Total as Total,
    
    All_Modes as All_Modes,
    Solo as Solo,
    Doubles as Doubles,
    Triples as Triples,
    Quadriples as Quadriples
)

from .utils import packages

from .utils import (
    Gamemodes as Gamemodes,
    Intervals as Intervals,
    Modes as Modes
)

from .utils import header

from .RefactorAPI.Guilds import Guild
from .RefactorAPI.Profiles import PikaProfile
from .RefactorAPI.Leaderboard import CombinedLeaderboard
from .RefactorAPI.NetworkStatus import NetworkStatus
from .RefactorAPI.Recaps import Recap

from .Games.Bedwars import Bedwars
from .Games.Skywars import Skywars
from .Games.Practice import Unrankedpractice, Rankedpractice

from .Punishments import History

from .Ratelimits import PikaNetworkRateLimitHandler as RateLimits

from .ResponseError import faulty

from ._Logger import log

from .utils import Check, Missing

from .utils import (
    APIBlockedException,
    APIResponseError,
    Do_Not_Touch
)

from typing import (
    Union, Literal, List, Tuple
)


class PikaAnnotations:
    """
    Helper class
    ~~~~~~~~
    =============

    - Represents `TypeAnnotation` for PikaNetwork's API wrapper.
    """
    Stats             =      Union[Bedwars, Skywars, Unrankedpractice, Rankedpractice]  |     None      | Missing
    Profile           =      PikaProfile                                                |     None      | Missing
    Guild             =      Guild                                                      |     None      | Missing
    Punishments       =      History                                                    |     None      | Missing
    Status            =      NetworkStatus                                              |     None      | Missing
    Recap             =      Recap                                                      |     None      | Missing

    Pikanetwork       :      'Pikanetwork'                             =     None, ..., None, ..., None | Missing
    PikaAnnotations   :      'PikaAnnotations'                         =     None, ..., None, ..., None | Missing


class Pikanetwork:
    """
    Pikanetwork API Wrapper
    ~~~~~~~~~~~~~~~~~~~~~~~
    A basic wrapper for the PikaNetwork's API.

    =============================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """

    def __init__(self, *, session: packages.aiohttp.ClientSession = ...) -> None:
        self.session = session
        self.cache = {...}

    async def __aenter__(self):
        if self.session == Ellipsis:
            self.session = packages.aiohttp.ClientSession(headers = await header())
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self.session and self.session != ...:
            await self.session.close()

    
    class PikaAnnotations:
        """
        Helper class
        ~~~~~~~~
        =============

        - Represents `TypeAnnotation` for PikaNetwork's API wrapper.
        """
        Stats             =      Union[Bedwars, Skywars, Unrankedpractice, Rankedpractice]  |     None      | Missing
        Profile           =      PikaProfile                                                |     None      | Missing
        Guild             =      Guild                                                      |     None      | Missing
        Punishments       =      History                                                    |     None      | Missing
        Status            =      NetworkStatus                                              |     None      | Missing
        Recap             =      Recap                                                      |     None      | Missing

        Pikanetwork       :      'Pikanetwork'                             =     None, ..., None, ..., None | Missing
        PikaAnnotations   :      'PikaAnnotations'                         =     None, ..., None, ..., None | Missing


    async def Profile(
            self,
            player: str,
            **kwargs: Missing
            ) -> Union[PikaProfile, None]:
        """
        :Profile API: Fetches profile data of a user.
        

        Parameters:
            - player (str) : In game name of the player.
            
        Returns:
            - If player is found returns `PikaProfile`, which can be further extracted.
            - `None` if unable to find player.

        Raise:
            - `APIBlockedException` if request is getting blocked.
            - `APIResponseError` if API isn't responding.
        
        
        :Example: https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/Profile.py 
        """

        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config

        await RateLimits.avoid_rate_limits()
        async with self.session.get(f'https://stats.pika-network.net/api/profile/{player}') as resp:

            status = resp.status

            if status == 200:
                return PikaProfile(await resp.json(), session=self.session)

            elif status == 429 and Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Profile(player)

            elif status == 400 or status == 204 or status == 404:
                packages.asyncio.create_task(log(player, ' not found!'))
                return None
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass
   

    async def Stats(
            self,
            player: str,
            gamemode: Literal["bedwars", "skywars", "unrankedpractice", "rankedpractice"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"],
            **kwargs: Missing
            ) -> Union[Bedwars, Skywars, Unrankedpractice, Rankedpractice, None]:
        """
        Stats API
        ~~~~~~~~~

        Parameters:
            - player (str) : Ingame name of the player.
            - gamemode (Literal) : minigame of the stats.
            - interval (Literal) : Timespan of the stats.
            - mode (Literal) : mode of the minigames.
            
        Returns:
            - If player is found returns minigame.
            - Returns None if unable to find player.

        Raise:
            - `ValueError` if gamemode, interval or mode is invalid.
            - `APIBlockedException` if request is getting blocked.
            - `APIResponseError` if API isn't responding.
        
        :Example: https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/Stats.py
        """
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config

        if Recursion <= 0:
            player, gamemode, interval, mode = str(player), gamemode.lower(), interval.lower(), mode.upper()
            Check().__stats_input__(gamemode, interval, mode)

        await RateLimits.avoid_rate_limits()
        async with self.session.get(f'https://stats.pika-network.net/api/profile/{player}/leaderboard?type={gamemode}&interval={interval}&mode={mode}') as resp:

            status = resp.status

            if status == 200:
                data: dict = await resp.json()

                if await faulty(data) and Recursion <= config.allowed_recursion:
                    Recursion += 1
                    packages.asyncio.create_task(log('Faulty Stats detected: ', Recursion, 'X'))
                    return await self.Stats(player, gamemode, interval, mode, Recursion=Recursion)

                return {
                    'bedwars': Bedwars,
                    'skywars': Skywars,
                    'unrankedpractice': Unrankedpractice,
                    'rankedpractice': Rankedpractice
                }.get(
                    gamemode,
                    lambda x: None
                )(data)


            elif status == 429 and Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Stats(player, gamemode, interval, mode, Recursion=Recursion)
            
            elif status == 204:
                packages.asyncio.create_task(log(player, ' is hidden from the API!'))
                return None
            
            elif status == 400 or status == 404:
                packages.asyncio.create_task(log(player, ' not found!'))
                return None
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass


    async def Guild(
            self,
            guild: str,
            **kwargs: Missing
            ) -> Union[Guild, None]:
        """
        Guild API
        ~~~~~~~~~

        Parameters:
            - guild (str) : Name of the guild.
        
        Returns:
            - If guild is found returns `GuildContext`.
            - Returns `None` if unable to find guild.

        Raise:
            - `APIBlockedException` if request is getting blocked.
            - `APIResponseError` if API isn't responding.

        :Example: https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/Guild.py
        """
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config

        await RateLimits.avoid_rate_limits()
        async with self.session.get(f'https://stats.pika-network.net/api/clans/{guild}') as resp:

            status = resp.status

            if status == 200:
                return Guild(await resp.json())
            
            elif status == 429 and Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Guild(guild)
            
            elif status == 400 or status == 204  or status == 404:
                packages.asyncio.create_task(log(guild, ' not found!'))
                return None
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass


    async def __leaderboard__helper__(
            self,
            gamemode,
            stats,
            interval,
            mode,
            offset: int,
            limit: int,
            **kwargs
        ) -> Union[dict, None]:
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config
        await RateLimits.avoid_rate_limits()
        async with self.session.get(f'https://stats.pika-network.net/api/leaderboards?type={gamemode}&stat={stats}&interval={interval}&mode={mode}&offset={offset}&limit={limit}') as resp:
            
            status = resp.status

            if status == 200:
                return await resp.json()
            
            elif status == 429 and Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.__leaderboard__helper__(gamemode, stats, interval, mode, offset, limit, Recursion=Recursion)
            
            elif status == 400 or status == 204  or status == 404:
                return None
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass


    async def Leaderboard(
            self,
            gamemode: Literal["bedwars", "skywars", "unrankedpractice", "rankedpractice"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"],
            stats: Literal['wins', 'losses', 'kills', 'deaths', 'final_kills', 'final_deaths', 'bed_destroyed', 'melee_kills', 'void_kills', 'bow_kills', 'arrows_hit', 'arrows_shot', 'played', 'highest_win_streak', 'elo', 'projectile_kills', 'melee_dealt', 'melee_taken'],
            **kwargs: Missing
        ) -> CombinedLeaderboard | list[dict]:
        """
        Leaderboard API
        ~~~~~~~~~~

        Parameters:
            - gamemode (Literal) :  Minigame name.
            - interval (Literal) : Timespan of the stats.
            - mode (Literal) : mode of the minigames.
            - stats (Literal) : type of leaderboard you want.
        
        Returns:
            - List

        Raise:
            - `APIBlockedException` if request is getting blocked.
            - `APIResponseError` if API isn't responding.
        """
        Check().__leaderboard_input__(gamemode, interval, mode, stats)
        Leaderboard_ = []
        offset = 0
        for _ in range(0, 14):
            Leaderboard_.append(self.__leaderboard__helper__(gamemode, stats, interval, mode, offset, 25, kwargs=kwargs))
            offset += 25

        return CombinedLeaderboard(await packages.asyncio.gather(*[task for task in Leaderboard_])).__Leaderboard__()

    
    async def Status(
            self,
            **kwargs: Missing
        ) -> NetworkStatus | None:
        """
        Network Status API
        ~~~~~~~~~~
        
        Returns:
        - 'PikaNetworkStatus'
        - None if failed to get status
        """
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config
        await RateLimits.avoid_rate_limits()
        async with self.session.get("https://api.craftigames.net/count/play.pika-network.net") as resp:
            if resp.status == 200:
                return NetworkStatus(packages.json.loads(await resp.text()))
            elif Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Status()
            else:
                return None

    
    async def Recap(
            self,
            key: str,
            **kwargs: Missing
        ) -> 'Recap':
        """
        Recap API
        ~~~~~~~~~~

        Parameters:
            - key (str) : recap id.
        
        Returns:
            - Recap (class) if the correct key (recap_id) has been passed.
            - None if the the key (recap_id) is invaild.

        Raise:
            - `APIBlockedException` if request is getting blocked.
            - `APIResponseError` if API isn't responding.
        """
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config
        await RateLimits.avoid_rate_limits()
        async with self.session.get(f"https://stats.pika-network.net/api/recaps/{key}") as resp:

            status = resp.status

            if status == 200:
                return Recap(await resp.json())
            
            elif status == 404 or status == 400 or status == 204:
                packages.asyncio.create_task(log(key, ': recap not found!'))
                return None
            
            elif status == 429 and Recursion <= config.allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Recap(key, Recursion=Recursion)
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass


    async def Punishment(
            self,
            player: str,
            **kwargs: Missing
        ) -> Union[History, None]:
        """
        Punishment
        ~~~~~~~~~

        Parameters:
            - player (str): username of the player.
            
        Returns:
            - History
            - None
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Pikanetwork
        import asyncio

        async def Example(player: str):
            
            async with Pikanetwork() as API:
            
                punishment = await API.Punishment(player)

            if punishment is None:
                return
            
            warnings = await punishment.warns()
            
            # and many more...

            print(wins, losses, losses_leaderboard, wlr, final_kills_leaderboard)
        
        asyncio.run(Example(player='AnyPlayer'))
        """
        Recursion = kwargs.get('Recursion', 0)
        RaiseError = kwargs.get('RaiseError', True)
        from .utils import config
        await RateLimits.avoid_rate_limits()
        async with self.session.get(f'https://pika-network.net/bans/search/{player}/') as resp:
            status = resp.status

            if status == 200:
                return History(await resp.text())
            
            elif status == 429 and Recursion <= config.allowed_recursion:

                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(config.delay_after_exceeding_ratelimit)
                return await self.Punishment(player)
            
            elif status == 403 and RaiseError:
                raise APIBlockedException()
            
            else:
                if RaiseError:
                    raise APIResponseError(status_code=status)
                else: pass

    # MultiProcessings
    
    async def MultiProfile(
            self,
            players: List[str],
            ) -> list[tuple[str, 'PikaProfile']]:
        """
        Fetch Multiple Player Profiles
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        Processes a list of player usernames and fetches their profiles asynchronously.

        Parameters:
            - players (List[str]) : A list of player usernames to fetch profiles for.

        Returns:
            - Array (List[tuple[str, PikaProfile]]) : A list of tuples, each containing a player's username 
            and their associated profile. If a player is not registered on PikaNetwork, their profile will be `None`.

        :NOTE:
            This function suppresses any errors encountered during profile retrieval, meaning 
            errors will not be raised, but they are not resolved either.

        Example:
            https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/MultiProfile.py
        """
        from .utils import config
        profile = []

        for i in range(0, len(players), config.batch_size):
            batch_members = players[i:i + config.batch_size]
            batch_member_stats = await packages.asyncio.gather(*(self.Profile(member, RaiseError=False) for member in batch_members))
            for member, member_stats in zip(batch_members, batch_member_stats):
                if member_stats is None:
                    profile.append((member, None))
                    continue

                profile.append((member, member_stats))

            await packages.asyncio.sleep(config.batch_delay)

        return profile
    

    async def MultiStats(
            self,
            players: List[str],
            gamemode: Literal["bedwars", "skywars", "unrankedpractice", "rankedpractice"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"]
            ) -> List[Tuple[str, Union[Bedwars, Skywars, Unrankedpractice, Rankedpractice, None]]]:
        """
        Fetch Multiple Player Stats
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~

        Retrieves statistics for a list of players across different game modes and intervals asynchronously.

        Parameters:
            - players (List[str]) : A list of player usernames to fetch stats for.
            - gamemode (Literal) : The specific minigame for which stats are retrieved (e.g., bedwars, skywars).
            - interval (Literal) : The timespan for which stats are retrieved (e.g., weekly, monthly).
            - mode (Literal) : The specific mode within the minigame (e.g., solo, doubles).

        Returns:
            - Array (List[Tuple[str, Union[Bedwars, Skywars, Unrankedpractice, Rankedpractice, None]]]) : A list of tuples, each containing a player's username and their associated stats. 
            If a player has their stats hidden from the API or has not registered on Pikanetwork, the stats will be `None`.
        
        :NOTE:
            This function suppresses any errors encountered during stats retrieval, meaning 
            errors will not be raised, but they are not resolved either.
        Example:
            https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/MultiStats.py
        """
        from .utils import config
        stats = []

        for i in range(0, len(players), config.batch_size):
            batch_members = players[i:i + config.batch_size]
            batch_member_stats = await packages.asyncio.gather(*(self.Stats(member, gamemode, interval, mode, RaiseError=False) for member in batch_members))
            for member, member_stats in zip(batch_members, batch_member_stats):
                if member_stats is None:
                    stats.append((member, None))
                    continue

                stats.append((member, member_stats))

            await packages.asyncio.sleep(config.batch_delay)

        return stats
    

    async def MultiGuilds(
            self,
            guilds: List[str],
            ) -> List[Tuple[str, 'Guild']]:
        """
        Fetch Multiple Guilds
        ~~~~~~~~~~~~~~~~~~~~~

        Retrieves data for a list of guilds asynchronously.

        Parameters:
            - guilds (List[str]) : A list of guild names to fetch data for.

        Returns:
            - Array (List[Tuple[str, 'Guild']]) : A list of tuples, each containing a guild's name 
            and its associated `Guild` class. If a guild name is invalid, the corresponding 
            `Guild` class will be `None`.

        :NOTE:
            This function suppresses any errors encountered during Guild retrieval, meaning 
            errors will not be raised, but they are not resolved either.

        Example:
            https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/MultiGuilds.py
        """
        from .utils import config
        guilds_ = []

        for i in range(0, len(guilds), config.batch_size):
            batch_guilds = guilds[i:i + config.batch_size]
            batch_guild_ = await packages.asyncio.gather(*(self.Guild(guild, RaiseError=False) for guild in batch_guilds))
            for guild, guild_api in zip(batch_guilds, batch_guild_):
                if guild_api is None:
                    guilds_.append((guild, None))
                    continue

                guilds_.append((guild, guild_api))

            await packages.asyncio.sleep(config.batch_delay)

        return guilds_
    

    async def MultiRecaps(
            self,
            ids: List[str]
        ) -> List[Tuple[str, 'Recap']]:
        """
        Fetch Multiple Recaps
        ~~~~~~~~~~~~~~~~~~~~~

        Retrieves recap data for a list of IDs/keys asynchronously.

        Parameters:
            - ids (List[str]) : A list of IDs/keys to fetch recap data for.

        Returns:
            - Array (List[Tuple[str, 'Recap']]) : A list of tuples, each containing an ID/key and its 
            associated `Recap` class. If an ID/key is invalid, the corresponding `Recap` class 
            will be `None`.

        :NOTE:
            This function suppresses any errors encountered during Guild retrieval, meaning 
            errors will not be raised, but they are not resolved either.

        Example:
            https://github.com/sti1ltyping/CraftiGames.py-Examples/blob/main/MultiRecaps.py
        """
        from .utils import config
        recaps = []

        for i in range(0, len(ids), config.batch_size):
            batch_keys = ids[i:i + config.batch_size]
            batch_keys_ = await packages.asyncio.gather(*(self.Recap(key, RaiseError=False) for key in batch_keys))
            for recap_id, recap_api in zip(batch_keys, batch_keys_):
                if recap_api is None:
                    recaps.append((recap_id, None))
                    continue

                recaps.append((recap_id, recap_api))

            await packages.asyncio.sleep(config.batch_delay)

        return recaps


"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from .utils import (
    Weekly as Weekly,
    Monthly as Monthly,
    Yearly as Yearly,
    Total as Total,
    
    All_Modes as All_Modes,
    Solo as Solo,
    Doubles as Doubles,
    Triples as Triples,
    Quadriples as Quadriples
)

from .utils import packages

from .utils import (
    Gamemodes as Gamemodes,
    Intervals as Intervals,
    Modes as Modes
)

from .utils import header

from .RefactorAPI.Guilds import Clan
from .RefactorAPI.Profiles import JartexProfile
from .RefactorAPI.Leaderboard import CombinedLeaderboard
from .RefactorAPI.NetworkStatus import NetworkStatus
from .RefactorAPI.Recaps import Recap

from .Games.Bedwars import Bedwars
from .Games.Skywars import Skywars

from .Punishments import History

from .Ratelimits import (
    JartexNetworkRateLimitHandler as jr_
)

from .ResponseError import faulty

from ._Logger import log

from .utils import config

from .utils import Check

from .utils import (
    APIBlockedException,
    APIResponseError,
    Do_Not_Touch
)

from typing import (
    Union, Literal, List, Tuple
)


allowed_recursion = config.allowed_recursion
batch_size = config.batch_size
batch_delay = config.batch_delay
delay_after_exceeding_ratelimit = config.delay_after_exceeding_ratelimit


class JartexAnnotations:
    """
    Helper class
    ~~~~~~~~
    =============

    - Represents `TypeAnnotation` for JartexNetwork's API wrapper.
    """
    Stats = Union[Bedwars, Skywars]
    Profile = JartexProfile
    Clan  = Clan
    Punishments = History
    Status = NetworkStatus
    Recap = Recap
    Jartexnetwork: 'Jartexnetwork' = None, ...
    JartexAnnotations: 'JartexAnnotations' = None, ...


class Jartexnetwork:
    """
    Jartexnetwork API Wrapper
    ~~~~~~~~~~~~~~~~~~~~~~~
    A basic wrapper for the JartexNetwork's API.

    =============================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """

    def __init__(self) -> None:
        self.session = None
        self.cache = {...}

    async def __aenter__(self):
        self.session = packages.aiohttp.ClientSession(headers = await header())
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self.session:
            await self.session.close()

    
    class JartexAnnotations:
        """
        Helper class
        ~~~~~~~~
        =============

        - Represents `TypeAnnotation` for JartexNetwork's API wrapper.
        """
        Stats = Union[Bedwars, Skywars]
        Profile = JartexProfile
        Clan  = Clan
        Punishments = History
        Status = NetworkStatus
        Recap = Recap
        Jartexnetwork: 'Jartexnetwork' = None, ...
        JartexAnnotations: 'JartexAnnotations' = None, ...


    async def Profile(
            self,
            player: str,
            *,
            Recursion: 'Do_Not_Touch' = 0
            ) -> Union[JartexProfile, None]:
        """
        Profile API
        ~~~~~~~~~

        Parameters:
            - player (str): Ingame name of the player.
            
            Returns:
            - If player is found returs minigame.
            - Retunrs None if unable to find player.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetwork, JartexAnnotations
        import asyncio


        AnyPlayer: str = ...

        async def example():

            async with Jartexnetwork() as api:

                profile = await api.Profile(AnyPlayer)
                
                # Additionally you can also get the skin (BYTES) of the player.
                skin = await profile.skin.Default.face()

            if profile is None:
                return print(AnyPlayer, "has not registered on JartexNetwork.")
            
            username: str = profile.username
            level: int = profile.level
            rank: str = profile.rank
            discord_verified: bool = profile.discord_verified
            
            # and many more...

            if discord_verified:
                print(AnyPlayer, 'has verified their discord account.')

            # Extract the guild from player's profile
            # Returns 'Guild' class if player is in a guild else returns None
            clan = profile.clan()  

            if clan:
                name = clan.name
                leader = clan.owner
                # and many more...

                print(f"{username} is a member of {name}. Leader of that owner is {owner}")

            print(username, level, rank)


        asyncio.run(example())
        """
        await jr_.avoid_rate_limits()
        async with self.session.get(f'https://stats.jartexnetwork.com/api/profile/{player}') as resp:

            status = resp.status

            if status == 200:
                return JartexProfile(await resp.json(), session=self.session)

            elif status == 429 and Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Profile(player)

            elif status == 400 or status == 204 or status == 404:
                packages.asyncio.create_task(log(player, ' not found!'))
                return None
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()
   

    async def Stats(
            self,
            player: str,
            gamemode: Literal["bedwars", "skywars"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"],
            *,
            Recursion: 'Do_Not_Touch' = 0
            ) -> Union[Bedwars, Skywars, None]:
        """
        Stats API
        ~~~~~~~~~

        Parameters:
            - player (str): Ingame name of the player.
            - gamemode (Literal): minigame of the stats.
            - interval (Literal): Timespan of the stats.
            - mode (Literal): mode of the minigames.
            
        Returns:
            - If player is found returns minigame.
            - Returns None if unable to find player.

        Raise:
            - Error if gamemode, interval or mode is invalid.
            - `APIBlockedException` if request is getting blocked.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetwork, JartexAnnotations
        import asyncio


        AnyPlayer: str = ...

        async def example():

            async with Jartexnetwork() as api:

                stats = await api.Stats(AnyPlayer, "bedwars", "total", "all_modes")
            
            if stats is None:
                return print("Player not found or Player is hidden from the API!")
            
            wins = stats.wins
            losses = stats.losses
            wlr = stats.wlr
            win_rate = stats.win_rate
            beds_destroyed = stats.beds_destroyed

            # and more...

            print(wins, losses, wlr, win_rate, beds_destroyed)

        asyncio.run(example())
        """
        if Recursion <= 0:
            player, gamemode, interval, mode = str(player), gamemode.lower(), interval.lower(), mode.upper()
            Check().__stats_input__(gamemode, interval, mode)

        await jr_.avoid_rate_limits()
        async with self.session.get(f'https://stats.jartexnetwork.com/api/profile/{player}/leaderboard?type={gamemode}&interval={interval}&mode={mode}') as resp:

            status = resp.status

            if status == 200:
                data: dict = await resp.json()

                if await faulty(data) and Recursion <= allowed_recursion:
                    Recursion += 1
                    packages.asyncio.create_task(log('Faulty Stats detected: ', Recursion, 'X'))
                    return await self.Stats(player, gamemode, interval, mode, Recursion=Recursion)

                return {
                    'bedwars': Bedwars,
                    'skywars': Skywars
                }.get(
                    gamemode,
                    lambda x: None
                )(data)


            elif status == 429 and Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Stats(player, gamemode, interval, mode, Recursion=Recursion)
            
            elif status == 204:
                packages.asyncio.create_task(log(player, ' is hidden from the API!'))
                return None
            
            elif status == 400 or status == 404:
                packages.asyncio.create_task(log(player, ' not found!'))
                return None
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()


    async def Clan(
            self,
            clan: str,
            *,
            Recursion: 'Do_Not_Touch' = 0
            ) -> Union[Clan, None]:
        """
        Clan API
        ~~~~~~~~~

        Parameters:
            - clan (str): Name of the clan.
            
            Returns:
            - If clan is found returns ClanContext.
            - Returns None if unable to find the clan.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetwork, JartexAnnotations
        import asyncio


        AnyClan: str = ...

        async def example():

            async with from Jartexnetwork() as api:

                clan = await api.Clan(AnyClan)
            
            if clan is None:
                return print("clan not found!")
            
            name = clan.name
            tag = clan.tag
            level = clan.level
            member_list = clan.member_list
            member_count = clan.member_count
            owner = clan.owner

            # and more...

            print(name, tag, level, member_count, owner)
            print(member_list)

        asyncio.run(example())
        """
        await jr_.avoid_rate_limits()
        async with self.session.get(f'https://stats.jartexnetwork.com/api/clans/{clan}') as resp:

            status = resp.status

            if status == 200:
                return Clan(await resp.json())
            
            elif status == 429 and Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Clan(clan)
            
            elif status == 400 or status == 204 or status == 404:
                packages.asyncio.create_task(log(clan, ' not found!'))
                return None
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()


    async def __leaderboard__helper__(
            self,
            gamemode,
            stats,
            interval,
            mode,
            offset: int,
            limit: int,
            *,
            Recursion: 'Do_Not_Touch' = 0
        ) -> Union[dict, None]:

        await jr_.avoid_rate_limits()
        async with self.session.get(f'https://stats.jartexnetwork.com/api/leaderboards?type={gamemode}&stat={stats}&interval={interval}&mode={mode}&offset={offset}&limit={limit}') as resp:
            
            status = resp.status

            if status == 200:
                return await resp.json()
            
            elif status == 429 and Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.__leaderboard__helper__(gamemode, stats, interval, mode, offset, limit, Recursion=Recursion)
            
            elif status == 400 or status == 204 or status == 404:
                return None
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()


    async def Leaderboard(
            self,
            gamemode: Literal["bedwars", "skywars"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"],
            stats: Literal['wins', 'losses', 'kills', 'deaths', 'final_kills', 'final_deaths', 'bed_destroyed', 'melee_kills', 'void_kills', 'bow_kills', 'arrows_hit', 'arrows_shot', 'played', 'highest_win_streak', 'projectile_kills']
        ) -> CombinedLeaderboard | list[dict]:
        """
        Leaderboard API
        ~~~~~~~~~~

        Parameters:
        - gamemode: minigame of the stats.
        - interval: Timespan of the stats.
        - mode: mode of the minigames.
        - stats: type of leaderboard you want.
        
        Returns:
        - List
        
        """
        Check().__leaderboard_input__(gamemode, interval, mode, stats)
        Leaderboard_ = []
        offset = 0
        for _ in range(0, 14):
            Leaderboard_.append(self.__leaderboard__helper__(gamemode, stats, interval, mode, offset, 25))
            offset += 25

        return CombinedLeaderboard(await packages.asyncio.gather(*[task for task in Leaderboard_])).__Leaderboard__()

    
    async def Status(
            self,
            *,
            Recursion: 'Do_Not_Touch' = 0
        ) -> NetworkStatus:
        """
        Network Status API
        ~~~~~~~~~~
        
        Returns:
        - 'NetworkStatus'
        """
        await jr_.avoid_rate_limits()
        async with self.session.get("https://api.craftigames.net/count/play.jartexnetwork.com") as resp:
            if resp.status == 200:
                return NetworkStatus(packages.json.loads(await resp.text()))
            elif Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Status()
            else:
                return None

    
    async def Recap(
            self,
            key: str,
            *,
            Recursion: 'Do_Not_Touch' = 0
        ) -> 'Recap':
        """
        Recap API
        ~~~~~~~~~~

        Parameters:
        - key (str): recap id.
        
        Returns:
        - Recap (class) if the correct key (recap_id) has been passed.
        - None if the the key (recap_id) is invaild.
        """
        await jr_.avoid_rate_limits()
        async with self.session.get(f"https://stats.jartexnetwork.com/api/recaps/{key}") as resp:

            status = resp.status

            if status == 200:
                return Recap(await resp.json())
            
            elif status == 404 or status == 400 or status == 204:
                packages.asyncio.create_task(log(key, ': recap not found!'))
                return None
            
            elif status == 429 and Recursion <= allowed_recursion:
                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Recap(key, Recursion=Recursion)
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()


    async def Punishment(
            self,
            player: str,
            *,
            Recursion: 'Do_Not_Touch' = 0
        ) -> Union[History, None]:
        """
        Punishment
        ~~~~~~~~~

        Parameters:
            - player (str): username of the player.
            
        Returns:
            - History
            - None
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetwork
        import asyncio

        async def Example(player: str):
            
            async with Jartexnetwork() as API:
            
                punishment = await API.Punishment(player)

            if punishment is None:
                return
            
            warnings = await punishment.warns()
            
            # and many more...

            print(wins, losses, losses_leaderboard, wlr, final_kills_leaderboard)
        
        asyncio.run(Example(player='AnyPlayer'))
        """
        await jr_.avoid_rate_limits()
        async with self.session.get(f'https://jartexnetwork.com/bans/search/{player}/') as resp:
            status = resp.status

            if status == 200:
                return History(await resp.text())
            
            elif status == 429 and Recursion <= allowed_recursion:

                Recursion += 1
                packages.asyncio.create_task(log('Exceeded ratelimit: ', Recursion, 'X'))
                await packages.asyncio.sleep(delay_after_exceeding_ratelimit)
                return await self.Punishment(player)
            
            elif status == 403:
                raise APIBlockedException()
            
            else:
                raise APIResponseError()

    # MultiProcessings

    async def MultiProfile(
            self,
            api: 'Jartexnetwork',
            players: list[str],
            ) -> list[tuple[str, 'JartexProfile']]:
        """
        Multiple Profiles
        ~~~~~~~~~~~~~~

        Parameters:
            - api: Instance of Jartexnetork class
            - players: List of players.
            
        Returns:
            - An array of players usernames and profile, it can further be used to extract specific items from profile.
            - Profile is None for any players that are not registered on Jartexnetork.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetwork, JartexAnnotations
        import asyncio


        List_of_Players: list = [...]

        async def example():

            async with Jartexnetork() as api:
                response = await api.MultiProfile(api, List_of_Players)
            
            for username, profile_api in response:
                profile_api: JartexAnnotations.Profile    # Annotations for profile api

                # Error handling
                if profile_api is None:
                    print(f"Couldn't find {username}")      # Player has not registered on Jartexnetork.
                    continue

                print(username, profile_api.rank)
                

        asyncio.run(example())
        """
        profile = []

        for i in range(0, len(players), batch_size):
            batch_members = players[i:i + batch_size]
            batch_member_stats = await packages.asyncio.gather(*(api.Profile(member) for member in batch_members))
            for member, member_stats in zip(batch_members, batch_member_stats):
                if member_stats is None:
                    profile.append((member, None))
                    continue

                profile.append((member, member_stats))

            await packages.asyncio.sleep(batch_delay)

        return profile
    

    async def MultiStats(
            self,
            api: 'Jartexnetwork',
            players: list[str],
            gamemode: Literal["bedwars", "skywars"],
            interval: Literal["weekly", "monthly", "yearly", "total"],
            mode: Literal["all_modes", "solo", "doubles", "triples", "quad"]
            ) -> list[tuple[str, Union[Bedwars, Skywars]]]:
        """
        Multiple stats
        ~~~~~~~~~~~~~~

        Parameters:
            - api: Instance of Jartexnetork class
            - players: List of players.
            - gamemode: Minigame of the stats.
            - interval: Timespan of the stats.
            - mode: Mode of the minigames.
            
        Returns:
            - An array of player's usernames and stats (Games class), it can further be used to extract specific stats.
            - stats (Games class) is None for the players hidden from the API.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetork, JartexAnnotations
        import asyncio


        List_of_Players: list = [...]

        async def example():

            async with Jartexnetork() as api:
                response = await api.Stats(api, List_of_Players, "total", "all_modes")
            
            for username, stats_api in response:
                stats_api: JartexAnnotations.Stats    # Annotations for stats api

                # Error handling
                if stats_api is None:
                    print(f"Couldn't find {username}")      # Maybe hidden from the api or incorrect username
                    continue

                print(username, stats_api.wins)
                

        asyncio.run(example())
        """
        stats = []

        for i in range(0, len(players), batch_size):
            batch_members = players[i:i + batch_size]
            batch_member_stats = await packages.asyncio.gather(*(api.Stats(member, gamemode, interval, mode) for member in batch_members))
            for member, member_stats in zip(batch_members, batch_member_stats):
                if member_stats is None:
                    stats.append((member, None))
                    continue

                stats.append((member, member_stats))

            await packages.asyncio.sleep(batch_delay)

        return stats
    

    async def MultiClans(
            self,
            api: 'Jartexnetwork',
            guilds: list[str],
            ) -> list[tuple[str, 'Clan']]:
        """
        Multiple Clans
        ~~~~~~~~~~~~~~

        Parameters:
            - api: Instance of Jartexnetork class.
            - clans: List of clans.
            
        Returns:
            - An array of clan's names and clan class, it can further be used to extract specific items from the clan class.
            - Clan is None for any invalid clan.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetork, JartexAnnotations
        import asyncio


        List_of_Clans: list = [...]

        async def example():

            async with Pikanetwork() as api:
                response = await api.MultiClans(api, List_of_Clans)
            
            for name, clan_api in response:
                clan_api: JartexAnnotations.Clan    # Annotations for clan api

                # Error handling
                if clan_api is None:
                    print(f"Couldn't find {name}")      # Incorrect clan name
                    continue

                print(name, clan_api.level)
                

        asyncio.run(example())
        """
        guilds_ = []

        for i in range(0, len(guilds), batch_size):
            batch_guilds = guilds[i:i + batch_size]
            batch_guild_ = await packages.asyncio.gather(*(api.Guild(guild) for guild in batch_guilds))
            for guild, guild_api in zip(batch_guilds, batch_guild_):
                if guild_api is None:
                    guilds_.append((guild, None))
                    continue

                guilds_.append((guild, guild_api))

            await packages.asyncio.sleep(batch_delay)

        return guilds_
    

    async def MultiRecaps(
            self,
            api: 'Jartexnetwork',
            ids: List[str]
        ) -> List[Tuple[str, 'Recap']]:
        """
        Multiple Recaps
        ~~~~~~~~~~~~~~

        Parameters:
            - api: Instance of Jartexnetork class
            - ids: List of ids/keys.
            
        Returns:
            - An array of id and Recap class, it can further be used to extract specific items from the Recap class.
            - Recap is None for any invalid id/key.
        
        Example
        ~~~~~~

        ~~~
        from CraftiGames import Jartexnetork, JartexnAnnotations
        import asyncio


        Ids_of_recaps: list = [...]

        async def example():

            async with Jartexnetork() as api:
                response = await api.MultiRecaps(api, Ids_of_recaps)
            
            for id, recap_api in response:
                recap_api: JartexAnnotations.Recap    # Annotations for recap api

                # Error handling
                if recap_api is None:
                    print(f"Couldn't find any recap of id: {id}")      # Incorrect id/key
                    continue

                print(id, recap_api.winners)
                

        asyncio.run(example())
        """
        recaps = []

        for i in range(0, len(ids), batch_size):
            batch_keys = ids[i:i + batch_size]
            batch_keys_ = await packages.asyncio.gather(*(api.Recap(key) for key in batch_keys))
            for recap_id, recap_api in zip(batch_keys, batch_keys_):
                if recap_api is None:
                    recaps.append((recap_id, None))
                    continue

                recaps.append((recap_id, recap_api))

            await packages.asyncio.sleep(batch_delay)

        return recaps


"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

class packages:
    import aiohttp
    import json
    import asyncio
    import time
    from datetime import (
        datetime,
        timedelta
    )
    from bs4 import BeautifulSoup
    from io import BytesIO
    import traceback
    import random
    from typing import (
        Literal, Union, List, AnyStr
    )
    from dateutil import parser
    import configparser
    import colorama
    from pathlib import Path


class Weekly:
    """Represents weekly interval."""
    def __str__(self) -> packages.Literal["weekly"]: return "weekly"
    
class Monthly:
    """Represents monthly interval."""
    def __str__(self) -> packages.Literal["monthly"]: return "monthly"
    
class Yearly:
    """Represents yearly interval."""
    def __str__(self) -> packages.Literal["yearly"]: return "yearly"
    
class Total:
    """Represents all times."""
    def __str__(self) -> packages.Literal["total"]: return "total"

class All_Modes:
    """Represents all_modes."""
    def __str__(self) -> packages.Literal["all_modes"]: return "all_modes"
    
class Solo:
    """Represents solo gamemode."""
    def __str__(self) -> packages.Literal["solo"]: return "solo"
    
class Doubles:
    """Represents doubles mode."""
    def __str__(self) -> packages.Literal["doubles"]: return "doubles"
    
class Triples:
    """Represents triples mode."""
    def __str__(self) -> packages.Literal["triples"]: return "triples"
    
class Quadriples:
    """Represents quadriples modes."""
    def __str__(self) -> packages.Literal["quad"]: return "quad"

class Gamemodes:

    """List of available gamemodes"""
    
    @property
    def bedwars() -> packages.Literal['bedwars']: return "bedwars"
    
    @property
    def skywars() -> packages.Literal['skywars']: return "skywars"
    
    @property
    def unrankedpractice() -> packages.Literal['unrankedpractice']: return "unrankedpractice"
    
    @property
    def rankedpractice() -> packages.Literal['rankedpractice']: return "rankedpractice"
    
    def all() -> packages.List[str]: return ['bedwars', 'skywars', 'unrankedpractice', 'rankedpractice']
    
class Intervals:

    """List of available intervals"""
    
    @property
    def weekly() -> packages.Literal["weekly"]: return "weekly"

    @property
    def monthly() -> packages.Literal["monthly"]: return "monthly"

    @property
    def yearly() -> packages.Literal["yearly"]: return "yearly"
    
    @property
    def total() -> packages.Literal["total"]: return "total"
    
    def all() -> packages.List[str]: return ['weekly', 'monthly', 'yearly', 'total']

class Modes:

    """List of available modes"""
    
    @property
    def all_modes():
        return "ALL_MODES"
    
    @property
    def solo():
        return "SOLO"
    
    @property
    def doubles():
        return "DOUBLES"
    
    @property
    def triples():
        return "TRIPLES"
    
    @property
    def quad():
        return "QUAD"
    
    def all() -> list:
        return ['ALL_MODES', 'SOLO', 'DOUBLES', 'TRIPLES', 'QUAD']
    

class APIResponseError(Exception):
    """Raise error on API's host side issue."""
    def __init__(self, status_code: int, message="Error!"):
        self.message = f"{status_code}: " + message
        super().__init__(self.message)

class APIBlockedException(Exception):
    """Exception raised when the API requests are blocked."""
    def __init__(self, message="403: API requests are being blocked."):
        self.message = message
        super().__init__(self.message)
    
class Do_Not_Touch: ...

class UnixTimestamp: ...

class Missing:

    def __str__(self) -> str:
        return  'Missing Arguments!'

class aiohttpClientHeader:
    user_agents = [
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Firefox/97.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.15; rv:97.0) Gecko/20100101 Firefox/97.0',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Edge/97.0.1072.62 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:91.0) Gecko/20100101 Firefox/91.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 11_2_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/89.0.4389.114 Safari/537.36',
        'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:91.0) Gecko/20100101 Firefox/91.0',
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/88.0.4324.150 Safari/537.36',
        'Mozilla/5.0 (iPhone; CPU iPhone OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Mobile/15E148 Safari/604.1',
        'Mozilla/5.0 (iPad; CPU OS 14_6 like Mac OS X) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0 Mobile/15E148 Safari/604.1',
        'Mozilla/5.0 (Android 10; Mobile; rv:91.0) Gecko/91.0 Firefox/91.0',
        'Mozilla/5.0 (Linux; Android 10; SM-G975F) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.120 Mobile Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.70 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_1) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.3 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/75.0.3770.142 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/14.0.2 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.109 Safari/537.36',
        'Mozilla/5.0 (X11; Ubuntu; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.121 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/77.0.3865.90 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/66.0.3359.181 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/71.0.3578.98 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_3) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.0.5 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36',
        'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14; rv:60.0) Gecko/20100101 Firefox/60.0',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14.6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/13.1.2 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.14.4) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.1 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/56.0.2924.87 Safari/537.36',
        'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/70.0.3538.77 Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_0) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/69.0.3497.100 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/61.0.3163.100 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/72.0.3626.109 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_2) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.1 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/12.0 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/67.0.3396.87 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/605.1.15 (KHTML, like Gecko) Version/11.1.2 Safari/605.1.15',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.99 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/63.0.3239.132 Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_5) AppleWebKit/601.7.7 (KHTML, like Gecko) Version/9.1.2 Safari/601.7.7',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/50.0.2661.94 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.111 Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.152 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/40.0.2214.93 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/36.0.1985.125 Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.153 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_8_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/34.0.1847.137 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/33.0.1750.152 Safari/537.36',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (X11; Linux x86_64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/32.0.1700.107 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/31.0.1650.63 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:52.0) Gecko/20100101 Firefox/52.0',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_5_8) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/30.0.1599.66 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64; rv:45.0) Gecko/20100101 Firefox/45.0',
        'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; AS; rv:11.0) like Gecko',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.4.11) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/29.0.1547.76 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/28.0.1500.95 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.107 Safari/537.36',
        'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/93.0.4577.82 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_14_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/94.0.4606.71 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/95.0.4638.54 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.45 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/97.0.4692.71 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_12_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/98.0.4758.102 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/99.0.4844.51 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.75 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.41 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.61 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/103.0.5060.114 Safari/537.36',
        'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.79 Safari/537.36',
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/105.0.5173.13 Safari/537.36'
    ]

    device_ids = [
        '0123456789abcdef',
        '9876543210abcdef',
        'abcdef0123456789',
        '1234567890abcdef',
        '4567890123abcdef',
        'feedfacecafebeef',
        'deadbeef12345678',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '9b8c7d6e5f4a3b2c',
        'c3b2a1d4e5f6g7h8',
        '7h8g6f5e4d3c2b1a',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '9b8c7d6e5f4a3b2c',
        'c3b2a1d4e5f6g7h8',
        '7h8g6f5e4d3c2b1a',
        '8a9b7c6d5e4f3g2h',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '9b8c7d6e5f4a3b2c',
        'c3b2a1d4e5f6g7h8',
        '7h8g6f5e4d3c2b1a',
        '8a9b7c6d5e4f3g2h',
        '9a8b7c6d5e4f3g2h',
        '0123456789abcdef',
        '9876543210abcdef',
        'abcdef0123456789',
        '1234567890abcdef',
        '4567890123abcdef',
        'feedfacecafebeef',
        'deadbeef12345678',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '9b8c7d6e5f4a3b2c',
        'c3b2a1d4e5f6g7h8',
        '7h8g6f5e4d3c2b1a',
        '8a9b7c6d5e4f3g2h',
        '1122334455667788',
        '99aabbccddeeff00',
        '0011223344556677',
        '8899aabbccddeeff',
        'ffeeddccbbaa9988',
        '7766554433221100',
        '0102030405060708',
        'a1b2c3d4e5f6g7h8',
        '1a2b3c4d5e6f7g8h',
        'f1e2d3c4b5a69788',
        '98a7b6c5d4e3f2f1',
        '0f1e2d3c4b5a6978',
        '9b8c7d6e5f4a3b2c'
    ]


async def header() -> aiohttpClientHeader:
    headers = {
        'User-Agent': packages.random.choice(aiohttpClientHeader.user_agents),
        'Device-Id': packages.random.choice(aiohttpClientHeader.device_ids)
    }
    return headers


class Check:

    valid_modes = {
        'Bedwars': [
            'all_modes',
            'solo',
            'doubles',
            'triples',
            'quad'
        ],

        'Skywars': [
            'all_modes',
            'solo',
            'doubles'
        ],

        'Practice': [
            'all_modes'
        ]
    }

    allowed_stats = {
        'Bedwars': [
            'wins',
            'losses',

            'kills',
            'deaths',

            'final_kills',
            'final_deaths',
            'bed_destroyed',

            'melee_kills',
            'void_kills',

            'bow_kills',
            'arrows_hit',
            'arrows_shot',

            'played',
            'highest_win_streak',
        ],
        
        'Skywars': [
            'wins',
            'losses',

            'kills',
            'deaths',

            'melee_kills',
            'void_kills',

            'bow_kills',
            'arrows_hit',
            'arrows_shot',

            'played',
            'highest_win_streak',
        ],

        'UnrankedPractice': [
            'wins',
            'losses',

            'kills',

            'projectile_kills',
            'melee_dealt',
            'melee_taken',
            
            'void_kills',
            'played',
            'highest_win_streak',
        
        ],

        'RankedPractice': [
            'wins',
            'losses',

            'kills',

            'projectile_kills',
            'melee_dealt',
            'melee_taken',

            'void_kills',
            'played',
            'highest_win_streak',
            'elo'
        ]

    }
    def __init__(self) -> None:
        self.valid_modes_in_skywars = self.valid_modes.get('Skywars')
        self.valid_modes_in_practice = self.valid_modes.get('Practice')

        self.Bedwars = self.allowed_stats.get('Bedwars')
        self.Skywars = self.allowed_stats.get('Skywars')
        self.UnrankedPractice = self.allowed_stats.get('UnrankedPractice')
        self.RankedPractice = self.allowed_stats.get('RankedPractice')

    def __stats_input__(
            self,
            gamemode: str,
            interval: str,
            mode: str
            ) -> None:
        
        
        if gamemode not in Gamemodes.all():
            raise ValueError('Invalid gamemode has been passed!')
        elif interval not in Intervals.all():
            raise ValueError('Invalid interval has been passed!')
        elif mode not in Modes.all():
            raise ValueError('Invalid mode has been passed!')
        
        if gamemode.lower() == 'skywars' and mode.lower() not in self.valid_modes_in_skywars:
            raise ValueError("Invalid mode, can't use that in skywars. Can only use solo or doubles.")
        elif gamemode.lower() in ['rankedpractice', 'unrankedpractice'] and mode.lower() not in self.valid_modes_in_practice:
            raise ValueError("Invalid mode, can't use that in practice. Can only use all_modes.")
        
    def __leaderboard_input__(
            self,
            gamemode: str,
            interval: str,
            mode: str,
            stats: str
            ) -> None:
        if gamemode.lower() not in Gamemodes.all():
            raise ValueError('Invalid gamemode has been passed!')
        elif interval.lower() not in Intervals.all():
            raise ValueError('Invalid interval has been passed!')
        elif mode.upper() not in Modes.all():
            raise ValueError('Invalid mode has been passed!')
        elif stats.lower() not in self.Bedwars and stats not in self.Skywars and stats not in self.UnrankedPractice and stats not in self.RankedPractice:
            raise ValueError('Invalid stats has been passed!')
        
        if gamemode.lower() == 'skywars' and mode.lower() not in self.valid_modes_in_skywars:
            raise ValueError("Invalid mode, can't use that in skywars. Can only use solo or doubles.")
        elif gamemode.lower() in ['rankedpractice', 'unrankedpractice'] and mode.lower() not in self.valid_modes_in_practice:
            raise ValueError("Invalid mode, can't use that in practice. Can only use all_modes.")
        
        if gamemode.lower() == 'bedwars' and stats.lower() not in self.Bedwars:
            raise ValueError("Invalid stats, you can't use that in bedars.")
        elif gamemode.lower() == 'skywars' and stats.lower() not in self.Skywars:
            raise ValueError("Invalid stats, you can't use that in skywars.")
        elif gamemode.lower() == 'rankedpractice' and stats.lower() not in self.RankedPractice:
            raise ValueError("Invalid stats, you can't use that in rankedpractice.")
        elif gamemode.lower() == 'unrankedpractice' and stats.lower() not in self.UnrankedPractice:
            raise ValueError("Invalid stats, you can't use that in unrankedpractice.")


class Configuration:
    def __init__(self, *, config_path='CraftiGames/settings.ini'):
        self.config_path = packages.Path(config_path)
        self._config = packages.configparser.ConfigParser()
        self._load_config()

    def _load_config(self):
        """Load the configuration from the file."""
        self._config.read(self.config_path)

        self.logging = self._config.getboolean('DEFAULT', 'Logging', fallback=False)
        self.allowed_recursion = self._config.getint('DEFAULT', 'Allowed_Recursion', fallback=30)
        self.batch_size = self._config.getint('DEFAULT', 'Batch_size', fallback=10)
        self.default_skins = self._config.get('Skins', 'Default', fallback="XSteve").split(',')

        self.delay_after_exceeding_ratelimit = self._config.getfloat('Ratelimit', 'Delay_After_Exceeding_Ratelimit', fallback=1.5)
        self.batch_delay = self._config.getfloat('Ratelimit', 'Delay_Between_Batches', fallback=1)
        self.interval = self._config.getint('Ratelimit', 'Interval', fallback=1)
        self.max_requests_per_interval = self._config.getint('Ratelimit', 'Request_Per_Interval', fallback=20)
        self.delay = self._config.getfloat('Ratelimit', 'Delay', fallback=1.5)

    def update(
        self,
        *,
        Logging: bool = False,
        Allowed_Recursion: int = 30,
        Batch_Size: int = 10,
        Batch_Delay: float = 1,
        Default_Skins: list = ['XSteve'],
        Interval: int = 1,
        Max_Requests_Per_Interval: int = 20,
        Delay: float = 1.5,
        Delay_After_Exceeding_Ratelimit: float = 1.5
    ) -> None:
        """
        Caution
        ~~~~
        Using this function is not recommended.

        _______________________________________________________________________________________________

        This function can be used to configure `CraftiGames` settings. [Dynamically updates config]

        NOTE:
        - In case you want to undo your change use the following code:

        `from CraftiGames import config`

        `config.reset() # or config.update() # sets everything to default`
        
        """
        _config = packages.configparser.ConfigParser()

        _config['DEFAULT'] = {
            'Logging': str(Logging),
            'Allowed_Recursion': str(Allowed_Recursion),
            'Batch_size': str(Batch_Size)
        }

        _config['Skins'] = {
            'Default': ', '.join(Default_Skins)
        }

        _config['Ratelimit'] = {
            'Delay_After_Exceeding_Ratelimit': str(Delay_After_Exceeding_Ratelimit),
            'Delay_Between_Batches': str(Batch_Delay),
            'Interval': str(Interval),
            'Request_Per_Interval': str(Max_Requests_Per_Interval),
            'Delay': str(Delay)
        }

        self.config_path.parent.mkdir(parents=True, exist_ok=True)

        with open(self.config_path, 'w', encoding='utf-8') as file:
            _config.write(file)

        self._load_config()

    def reset(self):
        """Reset configuration to default!"""
        self.update()

config = Configuration()


"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

class Stats:
    """
    Wraps Bedwars 
    ~~~~~

    ==================================================================================================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE."""

    def __init__(self, data: dict) -> None:
        self.raw: dict = data
    

    def _wins(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of wins and/or leaderboard position.
            
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
            
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        wins = self.raw.get("Wins", {}).get("entries", [{}])
        value_result = wins[0].get("value", "0") if wins and wins[0] else "0"
        leaderboard_result = wins[0].get("place", "0") if wins and wins[0] else "0"
            
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _losses(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of losses and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
            
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        Losses = self.raw.get("Losses", {}).get("entries", [{}])
        value_result = Losses[0].get("value", "0") if Losses and Losses[0] else "0"
        leaderboard_result = Losses[0].get("place", "0") if Losses and Losses[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _wlr(self) -> float:
        """
        Wins is to Losses ratio.

        Returns:
        - Ratio of wins is to losses upto 2 decimal places.
        """
        wins = self._wins(leaderboard=False)
        losse = self._losses(leaderboard=False)
        if losse == 0:
            return int(wins)
        else:
            ratio = int(wins) / int(losse)
            return float('{:.2f}'.format(ratio))


    def _win_rate(self) -> float:
        """
        Wins rate.

        Returns:
        - Win rate upto 2 decimal places.
        """
        wins = self._wins(leaderboard=False)
        losses = self._losses(leaderboard=False)
        total = wins + losses
        if total == 0 and wins == 0:
            return 0
        elif total == 0:
            return 100
        rate = wins/total * 100
        return float('{:.2f}'.format(rate))


    def _lose_rate(self) -> float:
        """
        Lose rate.
        
        Returns:
        - Lose rate upto 2 decimal places.
        """
        win_rate = self._win_rate()
        if win_rate == 0:
            return 0
        return 100 - win_rate


    def _kills(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of kills and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
            
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        kills = self.raw.get("Kills", {}).get("entries", [{}])
        value_result = kills[0].get("value", "0") if kills and kills[0] else "0"
        leaderboard_result = kills[0].get("place", "0") if kills and kills[0] else "0"
            
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _deaths(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of deaths and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
            
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        deaths = self.raw.get("Deaths", {}).get("entries", [{}])
        value_result = deaths[0].get("value", "0") if deaths and deaths[0] else "0"
        leaderboard_result = deaths[0].get("place", "0") if deaths and deaths[0] else "0"
            
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _kdr(self) -> float:
        """
        Kills is to Deaths ratio.

        Returns:
        - Ratio of kills is to deaths upto 2 decimal places.
        """
        kills = self._kills(leaderboard=False)
        deaths = self._deaths(leaderboard=False)
        if deaths == 0:
            return int(kills)
        else:
            ratio = int(kills) / int(deaths)
            return float('{:.2f}'.format(ratio))


    def _kill_rate(self) -> float:
        """
        Kill rate.

        Returns:
        - Kill rate upto 2 decimal places.
        """
        kills = self._kills(leaderboard=False)
        deaths = self._deaths(leaderboard=False)
        total = kills + deaths
        if total == 0 and kills == 0:
            return 0
        elif total == 0:
            return 100
        rate = kills/total * 100
        return float('{:.2f}'.format(rate))


    def _death_rate(self) -> float:
        """
        Death rate.
        
        Returns:
        - Death rate upto 2 decimal places.
        """
        kill_rate = self._kill_rate()
        if kill_rate == 0:
            return 0
        return 100 - kill_rate


    def _final_kills(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of final kills and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        Finals_deaths = self.raw.get("Final kills", {}).get("entries", [{}])
        value_result = Finals_deaths[0].get("value", "0") if Finals_deaths and Finals_deaths[0] else "0"
        leaderboard_result = Finals_deaths[0].get("place", "0") if Finals_deaths and Finals_deaths[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _final_deaths(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of final deaths and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        Finals = self.raw.get("Final deaths", {}).get("entries", [{}])
        value_result = Finals[0].get("value", "0") if Finals and Finals[0] else "0"
        leaderboard_result = Finals[0].get("place", "0") if Finals and Finals[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _fkdr(self) -> (int | float):
        """
        Final kills is to Final deaths ratio.

        Returns:
        - Ratio of final kills is to final deaths upto 2 decimal places.
        """
        final_kills = self._final_kills(leaderboard=False)
        final_deaths = self._final_deaths(leaderboard=False)
        if final_deaths == 0:
            return int(final_kills)
        else:
            ratio = int(final_kills) / int(final_deaths)
            return float('{:.2f}'.format(ratio))


    def _final_kill_rate(self) -> float:
        """
        Final kill rate.

        Returns:
        - Final kill rate upto 2 decimal places.
        """
        final_kills = self._final_kills(leaderboard=False)
        final_death = self._final_deaths(leaderboard=False)
        total = final_kills + final_death
        if total == 0 and final_kills == 0:
            return 0
        elif total == 0:
            return 100
        rate = final_kills/total * 100
        return float('{:.2f}'.format(rate))


    def _final_death_rate(self) -> float:
        """
        Final death rate.
        
        Returns:
        - Final death rate upto 2 decimal places.
        """
        final_kill_rate = self._final_kill_rate()
        if final_kill_rate == 0:
            return 0
        return 100 - final_kill_rate


    def _beds_destroyed(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of beds destroyed and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        beds_destroyed = self.raw.get("Beds destroyed", {}).get("entries", [{}])
        value_result = beds_destroyed[0].get("value", "0") if beds_destroyed and beds_destroyed[0] else "0"
        leaderboard_result = beds_destroyed[0].get("place", "0") if beds_destroyed and beds_destroyed[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _highest_winstreak_reached(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of heighest win streak and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        Winstreak = self.raw.get("Highest winstreak reached",{}).get("entries", [{}])
        value_result = Winstreak[0].get("value", "0") if Winstreak and Winstreak[0] else "0"
        leaderboard_result = Winstreak[0].get("place", "0") if Winstreak and Winstreak[0] else "0"

        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _bow_kills(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of bow kills and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        Bow_kills = self.raw.get("Bow_kills", {}).get("entries", [{}])
        value_result = Bow_kills[0].get("value", "0") if Bow_kills and Bow_kills[0] else "0"
        leaderboard_result = Bow_kills[0].get("place", "0") if Bow_kills and Bow_kills[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _arrows_shot(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of arrow shot and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        arrows_shot = self.raw.get("Arrows shot", {}).get("entries", [{}])
        value_result = arrows_shot[0].get("value", "0") if arrows_shot and arrows_shot[0] else "0"
        leaderboard_result = arrows_shot[0].get("place", "0") if arrows_shot and arrows_shot[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _arrows_hit(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of arrow hit and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        arrows_hit = self.raw.get("Arrows hit", {}).get("entries", [{}])
        value_result = arrows_hit[0].get("value", "0") if arrows_hit and arrows_hit[0] else "0"
        leaderboard_result = arrows_hit[0].get("place", "0") if arrows_hit and arrows_hit[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _ahr(self) -> float:
        """
        Arrow hit ratio.

        Returns:
        - Arrow hit is to arrow shot ratio upto 2 decimal places.
        """
        arrows_hit = self._arrows_hit(leaderboard=False)
        arrows_shot = self._arrows_shot(leaderboard=False)
        if arrows_shot == 0:
            return arrows_hit
        elif arrows_hit == 0:
            return 0
        ratio = arrows_hit/arrows_shot
        return float('{:.2f}'.format(ratio))

        
    def _arrow_hit_rate(self) -> float:
        """
        Arrow hit rate.

        Returns:
        - Arrow hit rate upto 2 decimal places.
        """
        arrows_hit = self._arrows_hit(leaderboard=False)
        arrows_shot = self._arrows_shot(leaderboard=False)
        total = arrows_hit + arrows_shot
        if total == 0 and arrows_hit == 0:
            return 0
        elif total == 0:
            return 100
        rate = arrows_hit/total * 100
        return float('{:.2f}'.format(rate))


    def _arrow_miss_rate(self) -> float:
        """
        Arrow miss rate.
        
        Returns:
        - Arrow miss rate upto 2 decimal places.
        """
        arrow_hit_rate = self._arrow_hit_rate()
        if arrow_hit_rate == 0:
            return 0
        return 100 - arrow_hit_rate
    

    def _hits_dealt(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of Hits dealt and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
            
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        hits_dealt = self.raw.get("Hits dealt", {}).get("entries", [{}])
        value_result = hits_dealt[0].get("value", "0") if hits_dealt and hits_dealt[0] else "0"
        leaderboard_result = hits_dealt[0].get("place", "0") if hits_dealt and hits_dealt[0] else "0"
            
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _hits_taken(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of hits taken and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        hits_taken = self.raw.get("Hits taken", {}).get("entries", [{}])
        value_result = hits_taken[0].get("value", "0") if hits_taken and hits_taken[0] else "0"
        leaderboard_result = hits_taken[0].get("place", "0") if hits_taken and hits_taken[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _hit_rate(self) -> float:
        """
        Hit rate.

        Returns:
        - Hit rate upto 2 decimal places.
        """
        hits_dealt = self._hits_dealt(leaderboard=False)
        hits_taken = self._hits_taken(leaderboard=False)
        total = hits_dealt + hits_taken
        if total == 0 and hits_dealt == 0:
            return 0
        elif total == 0:
            return 100
        rate = hits_taken/total * 100
        return float('{:.2f}'.format(rate))


    def _hit_taken_rate(self) -> float:
        """
        Hit taken rate.
        
        Returns:
        - Hit taken rate upto 2 decimal places.
        """
        hit_dealt_rate = self._hit_rate()
        if hit_dealt_rate == 0:
            return 0
        return 100 - hit_dealt_rate
    
    
    def _elo(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of Elo and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        elo = self.raw.get("Elo", {}).get("entries", [{}])
        value_result = elo[0].get("value", "0") if elo and elo[0] else "0"
        leaderboard_result = elo[0].get("place", "0") if elo and elo[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _melee_kills(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of melee kills and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        melee_kills = self.raw.get("Melee kills", {}).get("entries", [{}])
        value_result = melee_kills[0].get("value", "0") if melee_kills and melee_kills[0] else "0"
        leaderboard_result = melee_kills[0].get("place", "0") if melee_kills and melee_kills[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _void_kills(
        self,
        value: bool = True,
        leaderboard: bool = True
        ) -> (tuple[int, int] | int | None):
        """
        Get the number of void kills and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        void_kills = self.raw.get("Void kills", {}).get("entries", [{}])
        value_result = void_kills[0].get("value", "0") if void_kills and void_kills[0] else "0"
        leaderboard_result = void_kills[0].get("place", "0") if void_kills and void_kills[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _games_played(
            self,
            value: bool = True,
            leaderboard: bool = True
            ) -> (tuple[int, int] | int | None):
        """
        Get the number of games played and/or leaderboard position.
        
        Parameters:
        - value (bool): If True, return the number of wins.
        - leaderboard (bool): If True, return the leaderboard position.
        
        Returns:
        - If both value and leaderboard are True, returns a tuple of (value, leaderboard).
        - If only value is True, returns the number of wins.
        - If only leaderboard is True, returns the leaderboard position.
        - If both value and leaderboard are False, returns None.
        """
        games_played = self.raw.get("Games played", {}).get("entries", [{}])
        value_result = games_played[0].get("value", "0") if games_played and games_played[0] else "0"
        leaderboard_result = games_played[0].get("place", "0") if games_played and games_played[0] else "0"
        
        if value and leaderboard:
            return int(value_result), int(leaderboard_result)
        elif value:
            return int(value_result)
        elif leaderboard:
            return int(leaderboard_result)
        else:
            return None


    def _exploited(
            self
            ) -> int:
        """
        Number of  number of games left `/hub`.

        Returns:
        - Number of games left unplayed.
        """
        gamesplayed = self._games_played(leaderboard=False)
        wins = self._wins(leaderboard=False)
        losses = self._losses(leaderboard=False)

        return int(gamesplayed) - int(wins + losses)



"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from CraftiGames.utils import packages, UnixTimestamp
from dataclasses import dataclass

class player: ...
class value: ...

@dataclass
class stats:
    username: str
    Blocks_placed: int
    Resources_collected: int
    Blocks_moved: int
    Items_bought: int
    Deaths: int
    Final_kills: int
    Kills: int
    Beds_destroyed: int
    Assists: int
    Blocks_mined: int
    time_of_death: str


class TopPlayers:
    """
    Leaderboard (helper class)
    ~~~~~~~~

    ==================================================================================================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE."""
    
    def __init__(self, users: list[dict[str, dict[str, int]]]) -> None:
        self.users = users
        self.leaderboard = self.calculate_leaderboard()
    
    def calculate_leaderboard(self) -> dict[str, list[dict[str, int]]]:
        stats_leaderboard = {
            "time_of_death_ms": [],
            "blocks_moved": [],
            "items_bought": [],
            "resources_collected": [],
            "deaths": [],
            "blocks_placed": [],
            "assists": [],
            "final_kills": [],
            "blocks_mined": [],
            "kills": [],
            "beds_destroyed": []
        }
        
        for user in self.users:
            username = user.get("username", '')
            stats = user.get("stats", {})

            time_of_death_ms = int(stats.get("Time of death (ms)", 0))
            blocks_moved = int(stats.get("Blocks moved", 0))
            items_bought = int(stats.get("Items bought", 0))
            resources_collected = int(stats.get("Resources collected", 0))
            deaths = int(stats.get("Deaths", 0))
            blocks_placed = int(stats.get("Blocks placed", 0))
            assists = int(stats.get("Assists", 0))
            final_kills = int(stats.get("Final kills", 0))
            blocks_mined = int(stats.get("Blocks mined", 0))
            kills = int(stats.get("Kills", 0))
            beds_destroyed = int(stats.get("Beds destroyed", 0))

            stats_leaderboard["time_of_death_ms"].append({"player": username, "value": time_of_death_ms})
            stats_leaderboard["blocks_moved"].append({"player": username, "value": blocks_moved})
            stats_leaderboard["items_bought"].append({"player": username, "value": items_bought})
            stats_leaderboard["resources_collected"].append({"player": username, "value": resources_collected})
            stats_leaderboard["deaths"].append({"player": username, "value": deaths})
            stats_leaderboard["blocks_placed"].append({"player": username, "value": blocks_placed})
            stats_leaderboard["assists"].append({"player": username, "value": assists})
            stats_leaderboard["final_kills"].append({"player": username, "value": final_kills})
            stats_leaderboard["blocks_mined"].append({"player": username, "value": blocks_mined})
            stats_leaderboard["kills"].append({"player": username, "value": kills})
            stats_leaderboard["beds_destroyed"].append({"player": username, "value": beds_destroyed})
        
        for stat in stats_leaderboard:
            stats_leaderboard[stat] = sorted(stats_leaderboard[stat], key=lambda x: x["value"], reverse=True)
        
        return stats_leaderboard


class Recap:
    """
    Wrap Recap
    ~~~~~

    ==================================================================================================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE."""

    def __init__(self, data: dict):
        self.raw: dict = data
        self.top_players = TopPlayers(
            self.raw.get(
                "users", []
            )
        )

    @property
    def id(self) -> str:
        """
        Returns:
        - ID of the recap.
        """
        return self.raw.get("id", '')
    
    @property
    def link(self) -> str:
        """
        Returns:
        - PikaNetwork link to the recap.
        """
        return f"https://stats.pika-network.net/recap/{self.id}"
    
    @property
    def map_name(self) -> str:
        """
        Returns:
        - Name of the map.
        """
        return self.raw.get("mapName", '')
    
    @property
    def game_type(self) -> str:
        """
        Returns:
        - Gamemode of the recap i.e. 'bedwars', 'skywars,...
        """
        return self.raw.get("gameType", '')
    
    @property
    def game_server_name(self) -> str:
        """
        Returns:
        - Game server name.
        """
        return self.raw.get("gameServerName", '')
    
    @property
    def game_start(self) -> UnixTimestamp:
        """
        Returns:
        - Unix timestamp (usable in Discord timestamp) of game's starting date-time.
        """
        iso_format = self.raw.get("gameStart", None)
        if iso_format:
            try:
                dt = packages.datetime.fromisoformat(iso_format.replace("Z", "+00:00"))
                return int(dt.timestamp())
            except ValueError as e:
                return 0
        return 0
    
    @property
    def game_duration(self) -> str:
        """
        Returns:
        - Duration of game.
        """
        return self.raw.get("gameDuration", '')
    
    @property
    def players(self) -> list[str]:
        """
        Returns:
        - List of all players in the match.
        """
        _players = self.raw.get("users", [])
        return [player["username"] for player in _players] if _players else []
    
    @property
    def player_count(self) -> int:
        """
        Returns:
        - Number of players in the game.
        """
        return len(self.players)

    @property
    def winners(self) -> list[str]:
        """
        Returns:
        - List of the winners.
        """
        return self.raw.get("winners", [])
    
    @property
    def most_kills(self) -> int:
        """
        Returns:
        - Value of most kills in the game.
        """
        return self.top_players.leaderboard["kills"][0]["value"] if self.top_players.leaderboard["kills"] else 0
    
    @property
    def player_with_most_kills(self) -> str:
        """
        Returns:
        - Username of the player who has most kills.
        """
        return self.top_players.leaderboard["kills"][0]["player"] if self.top_players.leaderboard["kills"] else ""
    
    @property
    def most_deaths(self) -> int:
        """
        Returns:
        - Value of most deaths in the game.
        """
        return self.top_players.leaderboard["deaths"][0]["value"] if self.top_players.leaderboard["deaths"] else 0
    
    @property
    def player_with_most_deaths(self) -> str:
        """
        Returns:
        - Username of the player who has most deaths.
        """
        return self.top_players.leaderboard["deaths"][0]["player"] if self.top_players.leaderboard["deaths"] else ""

    @property
    def last_death_time_ms(self) -> int:
        """
        Returns:
        - Time (ms) when last player was killed.
        """
        return self.top_players.leaderboard["time_of_death_ms"][0]["value"] if self.top_players.leaderboard["time_of_death_ms"] else 0

    @property
    def last_player_to_die(self) -> str:
        """
        Returns:
        - Username of the last player to get killed.
        """
        return self.top_players.leaderboard["time_of_death_ms"][0]["player"] if self.top_players.leaderboard["time_of_death_ms"] else ""

    @property
    def most_blocks_moved(self) -> int:
        """
        Returns:
        - Value of most blocks moved in the game.
        """
        return self.top_players.leaderboard["blocks_moved"][0]["value"] if self.top_players.leaderboard["blocks_moved"] else 0

    @property
    def player_with_most_blocks_moved(self) -> str:
        """
        Returns:
        - Username of the player who has moved most blocks.
        """
        return self.top_players.leaderboard["blocks_moved"][0]["player"] if self.top_players.leaderboard["blocks_moved"] else ""

    @property
    def most_items_bought(self) -> int:
        """
        Returns:
        - Value of most items bought in the game.
        """
        return self.top_players.leaderboard["items_bought"][0]["value"] if self.top_players.leaderboard["items_bought"] else 0

    @property
    def player_with_most_items_bought(self) -> str:
        """
        Returns:
        - Username of the player who has bought most item.
        """
        return self.top_players.leaderboard["items_bought"][0]["player"] if self.top_players.leaderboard["items_bought"] else ""

    @property
    def most_resources_collected(self) -> int:
        """
        Returns:
        - Value of most resources collected in the game.
        """
        return self.top_players.leaderboard["resources_collected"][0]["value"] if self.top_players.leaderboard["resources_collected"] else 0

    @property
    def player_with_most_resources_collected(self) -> str:
        """
        Returns:
        - Username of the player who has collected most resources.
        """
        return self.top_players.leaderboard["resources_collected"][0]["player"] if self.top_players.leaderboard["resources_collected"] else ""

    @property
    def most_blocks_placed(self) -> int:
        """
        Returns:
        - Value of most blocks placed in the game.
        """
        return self.top_players.leaderboard["blocks_placed"][0]["value"] if self.top_players.leaderboard["blocks_placed"] else 0

    @property
    def player_with_most_blocks_placed(self) -> str:
        """
        Returns:
        - Username of the player who has placed most blocks.
        """
        return self.top_players.leaderboard["blocks_placed"][0]["player"] if self.top_players.leaderboard["blocks_placed"] else ""

    @property
    def most_assists(self) -> int:
        """
        Returns:
        - Value of most assists in the game.
        """
        return self.top_players.leaderboard["assists"][0]["value"] if self.top_players.leaderboard["assists"] else 0

    @property
    def player_with_most_assists(self) -> str:
        """
        Returns:
        - Username of the player who has assisted the most.
        """
        return self.top_players.leaderboard["assists"][0]["player"] if self.top_players.leaderboard["assists"] else ""

    @property
    def most_final_kills(self) -> int:
        """
        Returns:
        - Value of most final kills in the game.
        """
        return self.top_players.leaderboard["final_kills"][0]["value"] if self.top_players.leaderboard["final_kills"] else 0

    @property
    def player_with_most_final_kills(self) -> str:
        """
        Returns:
        - Username of the player who has most final kills.
        """
        return self.top_players.leaderboard["final_kills"][0]["player"] if self.top_players.leaderboard["final_kills"] else ""

    @property
    def most_blocks_mined(self) -> int:
        """
        Returns:
        - Value of most block mined in the game.
        """
        return self.top_players.leaderboard["blocks_mined"][0]["value"] if self.top_players.leaderboard["blocks_mined"] else 0

    @property
    def player_with_most_blocks_mined(self) -> str:
        """
        Returns:
        - Username of the player who has mined most blocks.
        """
        return self.top_players.leaderboard["blocks_mined"][0]["player"] if self.top_players.leaderboard["blocks_mined"] else ""

    @property
    def most_beds_destroyed(self) -> int:
        """
        Returns:
        - Value of most beds destroyed in the game.
        """
        return self.top_players.leaderboard["beds_destroyed"][0]["value"] if self.top_players.leaderboard["beds_destroyed"] else 0

    @property
    def player_with_most_beds_destroyed(self) -> str:
        """
        Returns:
        - Username of the player who has destroyed most beds.
        """
        return self.top_players.leaderboard["beds_destroyed"][0]["player"] if self.top_players.leaderboard["beds_destroyed"] else ""
    
    @property
    def leaderboard_time_of_death(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: time of death (ms).
        
        Returns:
        - A list of player and time of death sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 11142},
            {"player": "player_2", "value": 10523},
            {"player": "player_3", "value": 8523}
        ]
        ```
        """
        return self.top_players.leaderboard["time_of_death_ms"]
    
    @property
    def leaderboard_blocks_moved(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of blocks.
        
        Returns:
        - A list of player and number of blocks moved sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 53},
            {"player": "player_2", "value": 32},
            {"player": "player_3", "value": 20}
        ]
        ```
        """
        return self.top_players.leaderboard["blocks_moved"]
    
    @property
    def leaderboard_items_bought(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: items bought.
        
        Returns:
        - A list of player and number items bought moved sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 53},
            {"player": "player_2", "value": 32},
            {"player": "player_3", "value": 20}
        ]
        ```
        """
        return self.top_players.leaderboard["items_bought"]
    
    @property
    def leaderboard_resources_collected(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of resources collected.
        
        Returns:
        - A list of player and number of resources collected sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 53},
            {"player": "player_2", "value": 32},
            {"player": "player_3", "value": 20}
        ]
        ```
        """
        return self.top_players.leaderboard["resources_collected"]
    
    @property
    def leaderboard_deaths(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of deaths.
        
        Returns:
        - A list of player and number of deaths sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 8},
            {"player": "player_2", "value": 4},
            {"player": "player_3", "value": 2}
        ]
        ```
        """
        return self.top_players.leaderboard["deaths"]
    
    @property
    def leaderboard_blocks_placed(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of blocks placed.
        
        Returns:
        - A list of player and number of blocks placed sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 623},
            {"player": "player_2", "value": 332},
            {"player": "player_3", "value": 220}
        ]
        ```
        """
        return self.top_players.leaderboard["blocks_placed"]
    
    @property
    def leaderboard_assists(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of assists.
        
        Returns:
        - A list of player and number of assists sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 5},
            {"player": "player_2", "value": 2},
            {"player": "player_3", "value": 1}
        ]
        ```
        """
        return self.top_players.leaderboard["assists"]
    
    @property
    def leaderboard_final_kills(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of final kills.
        
        Returns:
        - A list of player and number of final kills sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 3},
            {"player": "player_2", "value": 2},
            {"player": "player_3", "value": 0}
        ]
        ```
        """
        return self.top_players.leaderboard["final_kills"]
    
    @property
    def leaderboard_blocks_mined(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of blocks mined.
        
        Returns:
        - A list of player and number of blocks mined sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 23},
            {"player": "player_2", "value": 32},
            {"player": "player_3", "value": 20}
        ]
        ```
        """
        return self.top_players.leaderboard["blocks_mined"]
    
    @property
    def leaderboard_kills(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of kills.
        
        Returns:
        - A list of player and number of kills sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 10},
            {"player": "player_2", "value": 5},
            {"player": "player_3", "value": 2}
        ]
        ```
        """
        return self.top_players.leaderboard["kills"]
    
    @property
    def leaderboard_beds_destroyed(self) -> list[dict[player, value]]:
        """
        Keys:
        - `player`: username of the player.
        - `value`: number of beds destroyed.
        
        Returns:
        - A list of player and number of beds destroyed sorted highest to lowest.

        Example:
        ```json
        [
            {"player": "player_1", "value": 5},
            {"player": "player_2", "value": 3},
            {"player": "player_3", "value": 0}
        ]
        ```
        """
        return self.top_players.leaderboard["beds_destroyed"]
    
    @property
    def player_stats(self):
        """
        Return
         - List of stats of players. (dataclass)
        """
        users: list[dict] = self.raw.get("users", [])
        _player_stats: list[stats] = []
        for _ in users:
            _stats: dict = _.get("stats", {})
            _player_stats.append(stats(
                username = _.get("username", None),
                Blocks_placed = int(_stats.get("Blocks placed", 0)),
                Resources_collected = int(_stats.get("Resources collected", 0)),
                Blocks_moved = int(_stats.get("Blocks moved", 0)),
                Items_bought = int(_stats.get("Items bought", 0)),
                Deaths = int(_stats.get("Deaths", 0)),
                Final_kills = int(_stats.get("Final kills", 0)),
                Kills = int(_stats.get("Kills", 0)),
                Beds_destroyed = int(_stats.get("Beds destroyed", 0)),
                Assists = int(_stats.get("Assists", 0)),
                Blocks_mined = int(_stats.get("Blocks mined", 0)),
                time_of_death = _stats.get("Time of death", None)
            )
        )
        return _player_stats


"""
MIT License

Copyright (c) 2024 sti1ltyping

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from CraftiGames.utils import packages, UnixTimestamp # type: ignore

class Guild:
    """
    Wraps guild API
    ~~~~~

    ==================================================================================================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """
        
    def __init__(self, data):
        self.raw: dict = data

    @property
    def name(self) -> str:
        """
        Returns:
        - Guild's name.
        """
        return self.raw.get("name", '')
        
    @property
    def tag(self) -> str:
        """
        Returns:
        - Guild's tag.
        """
        return self.raw.get("tag", '')
        
    @property
    def leader(self) -> str:
        """
        Returns:
        - Guild's leader's username.
        """
        return self.raw.get("owner", {}).get("username", '')

    @property
    def link(self) -> str:
        """
        Returns:
        - PikaNetwork link to the leader's profile.
        """
        return f"https://stats.pika-network.net/player/{self.leader}"
        
    @property
    def level(self) -> int:
        """
        Returns:
        - Guild's level.
        """
        return int(self.raw.get("leveling",{}).get("level", 0))
    
    @property
    def exp(self) -> int:
        """
        Returns:
        - Guild's exp.
        """
        return int(self.raw.get("leveling",{}).get("exp", 0))
    
    @property
    def totalexp(self) -> int:
        """
        Returns:
        - Guild's total exp.
        """
        return int(self.raw.get("leveling",{}).get("totalExp", 0))
        
    @property
    def member_list(self) -> list:
        """
        Returns:
        - Guild member's list.
        """    
        return [username["user"]["username"] for username in self.raw.get("members", [])]
        
    @property
    def member_count(self) -> int:
        """
        Returns:
        - Member count of the guild.
        """
        return len(self.member_list)
    
    @property
    def created_at(self) -> UnixTimestamp:
        """
        Returns:
        - UnixTimestamp (usable in discord timestamp) of guild's creation date-time.
        """
        return int(packages.datetime.timestamp(packages.datetime.fromisoformat(self.raw.get('creationTime', 0))))
    

class Clan:
    """
    Wraps Clan API
    ~~~~~

    ==================================================================================================

    MIT License

    Copyright (c) 2024 sti1ltyping

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.
    """
        
    def __init__(self, data):
        self.raw: dict = data

    @property
    def name(self) -> str:
        """
        Returns:
        - Clan's name.
        """
        return self.raw.get("name", '')
        
    @property
    def tag(self) -> str:
        """
        Returns:
        - Clan's tag.
        """
        return self.raw.get("tag", '')
    
    @property
    def trophies(self) -> int:
        """
        Returns:
        - Clan's current trophies.
        """
        return int(self.raw.get("currentTrophies", 0))
    
    @property
    def owner(self) -> str:
        """
        Returns:
        - Clan's owner's username.
        """
        return self.raw.get("owner", {}).get("username", '')

    @property
    def link(self) -> str:
        """
        Returns:
        - PikaNetwork link to the leader's profile.
        """
        return f"https://stats.jartexnetwork.com/player/{self.owner}"
        
    @property
    def level(self) -> int:
        """
        Returns:
        - Clan's level.
        """
        return int(self.raw.get("leveling",{}).get("level", 0))
    
    @property
    def exp(self) -> int:
        """
        Returns:
        - Clan's exp.
        """
        return int(self.raw.get("leveling",{}).get("exp", 0))
    
    @property
    def totalexp(self) -> int:
        """
        Returns:
        - Clan's total exp.
        """
        return int(self.raw.get("leveling",{}).get("totalExp", 0))
        
    @property
    def member_list(self) -> list:
        """
        Returns:
        - Clan member's list.
        """    
        return [username["user"]["username"] for username in self.raw.get("members", [])]
        
    @property
    def member_count(self) -> int:
        """
        Returns:
        - Clan count of the guild.
        """
        return len(self.member_list)
    
    @property
    def created_at(self) -> UnixTimestamp:
        """
        Returns:
        - UnixTimestamp (usable in discord timestamp) of Clan's creation date-time.
        """
        return int(packages.datetime.timestamp(packages.datetime.fromisoformat(self.raw.get('creationTime', 0))))